#!/usr/bin env python
import sys, os, time, re, getopt, pdb
from cov import Cov
from panel import Panel
from files import Files
import technicalreport, clinicalreport, designreport, covermiplot
#import testreport as clinicalreport

class CoverMiException(Exception):
    pass


def create_output_dir(output_path, bam_path):
    output_path = os.path.join(output_path, "{0}_covermi_output".format(os.path.splitext(os.path.basename(bam_path))[0]))
    try:
        os.mkdir(output_path)
    except OSError:
        if os.path.isdir(output_path):
            raise CoverMiException("{0} folder already exists".format(output_path))
        else:
            raise CoverMiException("Unable to create folder {0}".format(output_path))
    return output_path


def main(panel_path, bam_path, output_path, depth=None):

    panel = Panel(panel_path).load(bam_path=="")
    if depth is not None:
        panel["Options"]["Depth"] = int(depth)
    output_path = create_output_dir(output_path, bam_path)
    print "Processing..."

    if bam_path != "":
        bam_file_list = Files(bam_path, ".bam")
        if len(bam_file_list) == 1:
            clinical_report_path = output_path
            technical_report_path = output_path
        else:
            clinical_report_path = os.path.join(output_path, "clinical")
            technical_report_path = os.path.join(output_path, "technical")
            os.mkdir(clinical_report_path)
            os.mkdir(technical_report_path)

        output_stems = set([])
        for panel["Filenames"]["Run"], panel["Filenames"]["Sample"], path in bam_file_list:
            path += ".bam"
            start_time = time.time()
            print "{0}/{1}".format(panel["Filenames"]["Run"], panel["Filenames"]["Sample"])

            output_stem = panel["Filenames"]["Sample"]
            dup_num = 1
            while output_stem in output_stems:
                output_stem = "{0}({1})".format(panel["Filenames"]["Sample"], dup_num)
                dup_num += 1
            output_stems.add(output_stem)
            
            if "Amplicons" in panel:
                cov = Cov.load_bam(path, panel["Amplicons"], amplicons=True)
                technicalreport.create(cov.amplicon_info, panel, os.path.join(technical_report_path, output_stem))
            else:
                cov = Cov.load_bam(path, panel["Exons"], amplicons=False)
            
            clinicalreport.create(cov, panel, os.path.join(clinical_report_path, output_stem))
            covermiplot.plot(cov, panel, os.path.join(clinical_report_path, output_stem))
            seconds = int(time.time() - start_time)
            time_string = "{0} sec".format(seconds) if (seconds<60) else "{0} min {01} sec".format(seconds/60, seconds%60)
            print"file {0} of {1} completed in {2}".format(len(output_stems), len(bam_file_list), time_string)
    else:
        cov = Cov.perfect_coverage(panel["Amplicons"])
        designreport.create(cov, panel, os.path.join(output_path, panel["Filenames"]["Panel"]))
        covermiplot.plot(cov, panel, os.path.join(output_path, panel["Filenames"]["Panel"]))


if __name__ == "__main__":

    try:
        opts, args = getopt.getopt(sys.argv[1:], "p:b:o:", ["panel=", "bams=", "output="])
    except getopt.GetoptError as err:
        raise CoverMiException(str(err))

    output = None
    bams = None
    panel = None
    depth = None
    for o, a in opts:
        if o in ("-p", "--panel"):
            panel = a
        elif o in ("-b", "--bams"):
            bams = a
        elif o in ("-o", "--output"):
            output = a
        elif o in ("-d", "--depth"):
            depth = a
        else:
            raise CoverMiException("Unrecognised option {0}".format(o))

    main(panel, bams, output, depth)











# panel is a dict containing all of the data structures that define a panel
# 	Amplicons:		genomic range
# 	Exons:			genomic range
# 	Transcripts:		genomic range
# 	Depth:
#	Variants_Disease:	genomic range
#	Variants_Gene:		genomic range
#	Variants_Mutation:	genomic range
#	Filenames:		dict of all files in the panel directory with filetype as the key
#       Options:		dict of all options, including depth

# the following are only included for a design panel
#	 AllTranscripts: 	genomic range
#	 AllExons:		genomic range
#	 Excluded:		list of excluded amplicons


import os, re, pdb
from gr import Gr


class CoverMiException(Exception):
    pass


filetypes = ["Excluded", "Reference", "Depth", "Targets", "Targets", "Manifest", "Variants", 
             "Canonical", "Canonical", "DesignStudio", "Disease_Names", "SNPs", "VCF", "Options"]
regexps = ["^chr[1-9XYM][0-9]?:[0-9]+-[0-9]+\\s*\n", #Single column of amplicon names in the format chr1:12345-67890 - excluded
           "^.+?\t.+?\tchr.+?\t[+-]\t[0-9]+\t[0-9]+\t[0-9]+\t[0-9]+\t[0-9]+\t[0-9,]+\t[0-9,]+\\s*\n", #Eleven columns - refflat
           "^[0-9]+\\s*\n", #Single number - depth
           "^[a-zA-Z][a-zA-Z0-9-]*\\s*\n", #Single column - targets
           "^[a-zA-Z][a-zA-Z0-9-]* +[a-zA-Z0-9_]+\\s*\n", #Single column - targets
           "^\\[Header\\]", #Manifest
           "^.+?\t.+?\t.+?\t[a-zA-Z0-9]+\t(null|chr[1-9XYM][0-9]?)\t(null|[0-9]+)\t(null|[0-9]+)\t(null|[+-])\t", #Nine columns - variants
           "^[a-zA-Z][a-zA-Z0-9]*\t[a-zA-Z][a-zA-Z0-9-]* +[a-zA-Z0-9_]+\\s*\n", #Two columns - canonical
           "^[a-zA-Z][a-zA-Z0-9]*\t[a-zA-Z0-9_]+\\s*\n", #Two columns - canonical
           "^chr[0-9XYM]+\t[0-9]+\t[0-9]+\t[^\t]+\t[0-9]+\t[+-]\\s*\n", #Six column bedfile - design
           "^#Variants Disease Name Translation\\s*\n", #Disease_Names
           "^chr[1-9XYM][0-9]?:[0-9]+ [ATCG\.]+>[ATCG\.]\\s*\n", #SNPs
           "^##fileformat=VCFv[0-9\.]+\\s*\n", #VCF
           "#CoverMi options\\s*\n"] # CoverMi panel options


class Panel(dict):


    def __init__(self, panel_path):
        self.panel_path = panel_path
        for root, dirnames, filenames in os.walk(panel_path):
            for filename in filenames:
                filepath = os.path.join(root, filename)
                with file(filepath, "rU") as f:
                    testlines = [f.readline(), f.readline()]
            
                alreadyfound = ""
                for testline in testlines:
                    if not testline.endswith("\n"):
                        testline = testline+"\n"
                    for regexp, filetype in zip(regexps, filetypes):
                        if re.match(regexp, testline) is not None:
                            if alreadyfound != "":
                                raise CoverMiException("Not a valid CoverMi panel\nUnable to uniquely identify file {0}, matches both {1} and {2} file formats".format(filepath, filetype, alreadyfound))
                            if filetype in self:
                                raise CoverMiException("Not a valid CoverMi panel\nFiles {0} and {1} both match {2} file format".format(filepath, self[filetype], filetype))
                            alreadyfound = filetype
                            self[filetype] = filepath
                    if alreadyfound != "":
                        break
        return


    def read_options(self):
        options = {}
        if "Depth" in self:
            with file(self["Depth"], "rU") as f:
                options["Depth"] = f.read().strip()
        if "Options" in self:
            print "Loading options file: {0}".format(os.path.basename(self["Options"]))
            with file(self["Options"], "rU") as f:
                for line in f:
                    line = line.strip()
                    if line != "" and not line.startswith("#"):
                        line = line.split("=")
                        if len(line) != 2:
                            raise CoverMiException("Malformed options file: {0}".format("=".join(line)))
                        if line[0] in options:
                            raise CoverMiException("Duplicate option: {0}".format(line[0]))
                        options[line[0]] = line[1]
        if "Depth" in options:
            if  options["Depth"].isdigit():
                options["Depth"] = int(options["Depth"])
            else:
                raise CoverMiException("Depth is not numeric: {0}".format(panel["Options"]["Depth"]))
        return options


    def write_options(self, options):
        if "Options" not in self:
            self["Options"] = os.path.join(self.panel_path, "covermi_options.txt")
            if os.path.lexists(self["Options"]):
                raise CoverMiException("File covermi_options.txt exists but is not of correct format")
        with file(self["Options"], "wt") as f: 
            f.write("#CoverMi options\n")
            for key, value in options.iteritems():
                f.write("{0}={1}\n".format(key, value))
        if "Depth" in self:
            os.unlink(self["Depth"])


    def load(self, design=False):
        panel = {}    

        if "Reference" not in self:
            raise CoverMiException("Unable to identify a Reference file in {0}".format(self.panel_path))
        if design and "Manifest" not in self and "DesignStudio" not in self:
            raise CoverMiException("Unable to identify a Manifest file or a Design Studio Bedfile in {0}".format(self.panel_path))
        if "Manifest" in self and "DesignStudio" in self:
            raise CoverMiException("Both Manifest and Design Studio Files found in {0}".format(self.panel_path))

        panel["Options"] = self.read_options()

        if "Manifest" in self:
            print "Loading manifest file: {0}".format(os.path.basename(self["Manifest"]))
            panel["Amplicons"] = Gr.load_manifest(self["Manifest"])
            if "Excluded" in self:
                excluded = []
                with file(self["Excluded"], "rU") as f:
                    for line in f:
                        line = line.strip()
                        if line != "" and not line.startswith("#"):
                            excluded.append(line)
                if design:
                    panel["ExcludedAmplicons"] = panel("Amplicons").subset(excluded)
                panel["Amplicons"] = panel("Amplicons").subset(excluded, exclude=True)

        elif "DesignStudio" in self:
            print "Loading Design Studio amplicons bedfile: {0}".format(os.path.basename(self["DesignStudio"]))
            panel["Amplicons"] = Gr.load(self["DesignStudio"])
        elif "Targets" in self:
            print "No manifest or design studio bedfile in panel. Will perform coverage analysis over all exons of genes in targets file"
        else:
            raise CoverMiException("WARNING. No manifest or design studio bedfile or targeted genes file in panel. Aborting.")

        if "Reference" in self:
            print "Loading reference file: {0}".format(os.path.basename(self["Reference"]))
            if "Targets" in self:
                print "Loading targeted genes file: {0}".format(os.path.basename(self["Targets"]))
            if "Canonical" in self:
                print "Loading canonical gene list: {0}".format(os.path.basename(self["Canonical"]))
            else:
                print "WARNING. No canonical gene list found. Loading all transcripts of targeted genes"
            panel["Exons"], panel["Transcripts"] = Gr.load_refflat(self["Reference"], 
                                                                   file(self["Targets"], "rU") if ("Targets" in self) else None, 
                                                                   file(self["Canonical"], "rU") if ("Canonical" in self) else None)

            if "Targets" not in self and "Amplicons" in panel:
                print "WARNING. No file identifying targeted genes. All genes touched by an amplicon will be assumed to be of interest"
                panel["Transcripts"] = panel["Transcripts"].touched_by(panel["Amplicons"])
                panel["Exons"] = panel["Exons"].touched_by(panel["Transcripts"]).subset(panel["Transcripts"].names)
            if design:
                panel["AllExons"], panel["AllTranscripts"] = Gr.load_refflat(self["Reference"], None, file(self["Canonical"], "rU") if ("Canonical" in self) else "")

        if "Variants" in self:
            print "Loading variants file: {0}".format(os.path.basename(self["Variants"]))
            if "Disease_Names" in self:
                print "Loading disease names file: {0}".format(os.path.basename(self["Disease_Names"]))
            panel["Variants_Disease"] = Gr.load_variants(self["Variants"], "disease",
                                                         disease_names=file(self["Disease_Names"], "rU") if ("Disease_Names" in self) else None)
            panel["Variants_Gene"] = Gr.load_variants(self["Variants"], "gene", genes_of_interest=panel["Transcripts"].names, 
                                                      disease_names=file(self["Disease_Names"], "rU") if ("Disease_Names" in self) else None)
            panel["Variants_Mutation"] = Gr.load_variants(self["Variants"], "mutation", genes_of_interest=panel["Transcripts"].names,
                                                          disease_names=file(self["Disease_Names"], "rU") if ("Disease_Names" in self) else None)
            
        panel["Filenames"] = { "Panel" : os.path.basename(self.panel_path.rstrip(os.pathsep)) }
        for filetype in self:
            panel["Filenames"][filetype] = os.path.splitext(os.path.basename(self[filetype]))[0]
        
        return panel




#!/usr/bin env python
import sys, os, time, re, getopt, pdb
from cov import Cov
from panel import Panel
from files import Files
import clinicalreport, covermiplot

class CoverMiException(Exception):
    pass


def create_output_dir(output_path):
    output_path = os.path.join(output_path, "covermi_output")
    try:
        os.mkdir(output_path)
    except OSError:
        if os.path.isdir(output_path):
            raise CoverMiException("{0} folder already exists".format(output_path))
        else:
            raise CoverMiException("Unable to create folder {0}".format(output_path))
    return output_path


def main(panel_path, bamlist, output_path):
    panel = Panel(panel_path).load()
    output_path = create_output_dir(output_path)
    print "Processing..."
    if len(set([os.path.splitext(os.path.basename(fn))[0].split("_")[0] for fn, depths in bamlist])) > len(bamlist):
        print "Duplicate sample numbers. Quiting"
        exit()

    starttime = time.time()
    for bam_path, depths in bamlist:
        panel["Options"]["Depths"] = depths
        panel["Filenames"]["Sample"] = os.path.splitext(os.path.basename(bam_path))[0].split("_")[0]
        cov = Cov.load_bam(bam_path, panel["Exons"], amplicons=False)
        for depth in depths:
            panel["Options"]["Depth"] = depth
            clinicalreport.create(cov, panel, os.path.join(output_path, "{0}_{1}x_".format(panel["Filenames"]["Sample"], depth)))
            seconds = int(time.time() - starttime)
            time_string = "{0} sec".format(seconds) if (seconds<60) else "{0} min {01} sec".format(seconds/60, seconds%60)
            print "File {}, Depth {}, {}".format(panel["Filenames"]["Sample"], depth, time_string)
        panel["Options"]["Depth"] = max(depths)
        covermiplot.plot(cov, panel, os.path.join(output_path, panel["Filenames"]["Sample"]))


#    try:
#        opts, args = getopt.getopt(sys.argv[1:], "p:b:o:", ["panel=", "bams=", "output="])
#    except getopt.GetoptError as err:
#        raise CoverMiException(str(err))

#    output = None
#    bams = None
#    panel = None
#    depth = None
#    for o, a in opts:
#        if o in ("-p", "--panel"):
#            panel = a
#        elif o in ("-b", "--bams"):
#            bams = a
#        elif o in ("-o", "--output"):
#            output = a
#        elif o in ("-d", "--depth"):
#            depth = a
#        else:
#            raise CoverMiException("Unrecognised option {0}".format(o))

#    main(panel, bams, output, depth)











#
# Ilumina Manifest file format:
#    [Header]
#
#    [Probes]
#    0.  Name-ID
#    1.  Region-ID
#    2.  Target-ID 
#    3.  Species
#    4.  Build-ID
#    5.  Chromosome
#    6.  Start position (including primer, zero based)
#    7.  End position (including primer, zero based)
#    8.  Strand
#    9.  Upstream primer sequence (len() will give primer length)
#    10  Upstream hits
#    11. Downstream  primer sequence (len() will give primer length)
#s
#    [Targets]
#    0.  Target-ID
#    1.  Target-ID
#    2.  Target number (1 = on target, >1 = off target)
#    3.  Chromosome
#    4.  Start position (including primer, zero based)
#    5.  End position (including primer, zero based)
#    6.  Probe strand
#    7.  Sequence (in direction of probe strand)
# 
#
# Bedfile format 
#    0.  Chromosome
#    1.  Start (zero based)
#    2.  End   (one based)
#    3.  Name
#    4.  Score
#    5.  Strand
# 
#
#  Variant file format
#
# 
#
# 
#
# 
#

import copy, pdb

class CoverMiException(Exception):
    pass

class Entry(object):

    def __repr__(self):
        return "[\"{0}\", {1}, {2}, \"{3}\", \"{4}\"]".format(self.chrom, self.start, self.stop, self.name, self.strand)

    def __lt__(self, other):
        return self.start<other.start

    def __init__(self, chrom, start, stop, name, strand):
        self.name = name
        self.chrom = chrom
        self.start = int(start)
        self.stop = int(stop)
        self.strand = strand


class Gr(dict):

    KARYOTYPE = {"chr1":1, "chr2":2, "chr3":3, "chr4":4, "chr5":5, "chr6":6, "chr7":7, "chr8":8, "chr9":9, "chr10":10, "chr11":11, "chr12":12, 
                 "chr13":13, "chr14":14, "chr15":15, "chr16":16, "chr17":17, "chr18":18, "chr19":19, "chr20":20, "chr21":21, "chr22":22, "chrX":23, "chrY":24, "chrM":25}
    @classmethod
    def chromosome_number(genomicrange, chromosome): return genomicrange.KARYOTYPE[chromosome]
    MAX_CHR_LENGTH = 1000000000
    SPLICE_SITE_BUFFER = 5


    def __init__(self, entry=None):
        if entry is not None:
            self.construct(copy.copy(entry))


    @classmethod
    def load_bedfile(genomicrange, path):
        with file(path, "rU") as f:
            gr = genomicrange()
            for line in f:
                splitline = line.strip().split("\t")
                if splitline != [""]:
                    name = splitline[3] if len(splitline)>3 else "."
                    strand = splitline[5] if len(splitline)>5 else "."
                    gr.construct(Entry(splitline[0], 
                                        int(splitline[1])+1,
                                        splitline[2],
                                        name,
                                        strand))
        gr.sort()
        return gr


    @classmethod
    def load_manifest(genomicrange, path, ontarget=True, offtarget=False):
        with file(path, "rU") as f:
            gr = genomicrange()
            section = "Header"
            skip_column_names = False
            probes = {}
            rename_offtarget = {}

            for line in f:
                splitline = line.rstrip("\n").split("\t")

                if skip_column_names:
                    skip_column_names = False

                elif line[0] == "[":
                    section = line[1:].split("]")[0]
                    if section != "Header":
                            skip_column_names = True

                elif section == "Probes":
                    probes[splitline[2]] = (len(splitline[9]), len(splitline[11]))

                elif section == "Targets":
                    if splitline[2] == "1":
                        amplicon_name = "{0}:{1}-{2}".format(splitline[3], splitline[4], splitline[5])
                        rename_offtarget[splitline[0]] = amplicon_name
                    else:
                        amplicon_name = rename_offtarget[splitline[0]]
                    if ((splitline[2] == "1") and ontarget) or ((splitline[2] != "1") and offtarget):
                        gr.construct(Entry(splitline[3],
                                            int(splitline[4])+probes[splitline[0]][splitline[6]=="+"], 
                                            int(splitline[5])-probes[splitline[0]][splitline[6]=="-"], 
                                            amplicon_name, 
                                            splitline[6]))
        gr.sort()
        return gr


    @classmethod
    def load_refflat(genomicrange, path, genes_of_interest=None, canonical_transcripts=None):
        canonicaldict = {}
        if canonical_transcripts is not None:
            for line in canonical_transcripts:
                gene, transcript = line.strip().split("\t")
                canonicaldict[gene] = transcript.split(" ")[-1]

        needed = set([])
        doublecheck = {}
        include_everything = True
        if genes_of_interest is not None:
            for line in genes_of_interest:
                line = line.strip()
                if line != "":
                    include_everything = False
                    if line in canonicaldict:
                        needed.add(canonicaldict[line])
                        doublecheck[canonicaldict[line]] = line
                    else:
                        splitline = line.split()
                        if len(splitline) == 1:
                            needed.add(line)
                        elif len(splitline) == 2:
                            needed.add(splitline[1])
                            doublecheck[splitline[1]] = splitline[0]
                        else:
                            raise CoverMiException("Malformed line in gene list: {0}".format(line))
            
        with file(path, "rU") as f:
            multiple_copies = {}
            found = set([])
            exons = genomicrange()
            transcripts = genomicrange()
            for line in f:
                splitline = line.rstrip("\n").split("\t")
                if splitline[2] in genomicrange.KARYOTYPE:
                    if include_everything:
                        if splitline[0] in canonicaldict and splitline[1] != canonicaldict[splitline[0]]:
                            continue
                    elif splitline[0] in needed:
                        found.add(splitline[0])
                    elif splitline[1] in needed and splitline[0] == doublecheck[splitline[1]]:
                        found.add(splitline[1])
                    else:
                        continue

                    transcript = "{0} {1}".format(splitline[0], splitline[1])
                    name = transcript
                    if name not in multiple_copies:
                        multiple_copies[name] = 1
                    else:
                        if multiple_copies[name] == 1:
                            for gr in (exons, transcripts):
                                for entry in gr.all_entries:
                                    if entry.name == name:
                                        entry.name = "{0} copy 1".format(name)
                        multiple_copies[name] += 1
                        name = "{0} copy {1}".format(name, multiple_copies[name])
              
                    entry = Entry(splitline[2], int(splitline[4])+1-genomicrange.SPLICE_SITE_BUFFER, int(splitline[5])+genomicrange.SPLICE_SITE_BUFFER,  name, splitline[3])
                    entry.transcript = transcript
                    transcripts.construct(entry)

                    exon_numbers = range(1,int(splitline[8])+1) if (splitline[3] == "+") else range(int(splitline[8]),0,-1)            
                    for start, stop, exon in zip(splitline[9].rstrip(",").split(","), splitline[10].rstrip(",").split(","), exon_numbers):
                        entry = Entry(splitline[2], int(start)+1-genomicrange.SPLICE_SITE_BUFFER, int(stop)+genomicrange.SPLICE_SITE_BUFFER, name, splitline[3])
                        entry.exon = exon
                        exons.construct(entry)

        missing = needed - found
        if not include_everything and len(missing) > 0:
            missing = ["{0} {1}".format(doublecheck[trans], trans) if trans in doublecheck else trans for trans in missing]
            print "WARNING - {0} not found in reference file".format(", ".join(missing))        

        for gr in (exons, transcripts):
            gr.sort()

            #simplify gene names
            names = {}
            duplicates = set([])
            for entry in gr.all_entries:
                gene, transcript = entry.name.split(" ")[0:2]
                if gene in names and names[gene] != transcript:
                    duplicates.add(gene)
                else:
                    names[gene] = transcript
            for entry in gr.all_entries:
                splitgene = entry.name.split(" ")
                if splitgene[0] not in duplicates:
                    if len(splitgene) <= 2:
                        entry.name = splitgene[0]
                    else:
                        entry.name = splitgene[0]+" "+" ".join(splitgene[2:])

        return (exons, transcripts)


    @classmethod
    def load_variants(genomicrange, path, catagory, genes_of_interest=None, disease_names=None): 
        if genes_of_interest is not None:
            genes_of_interest = set([name.split()[0] for name in genes_of_interest])

        disease_dict = {}
        if disease_names is not None:
            for line in disease_names:
                if line[0] != "#":
                    line2 = line.rstrip().split("=")
                    if len(line2) == 2:
                        disease_dict[line2[0].strip("\" ")] = line2[1].strip("\" ")

        with file(path, "rU") as f:
            mutations = {}
            for line in f:
                splitline = line.rstrip().split("\t")
                if splitline == [""] or splitline[4] not in genomicrange.KARYOTYPE:
                    continue
                if genes_of_interest is not None and splitline[3] not in genes_of_interest:
                    continue
                mutation = "{0} {1} {2}:{3}-{4}".format(splitline[3], splitline[8], splitline[4], splitline[5], splitline[6])# gene mutation location
                disease = splitline[1].strip("\" ")
                if disease in disease_dict:
                    if disease_dict[disease] == "":
                        continue
                    else:
                        disease = disease_dict[disease]
                if catagory == "disease":
                    name = disease 
                    mutation = (mutation, name)
                elif catagory == "gene":
                    name = splitline[3]
                elif catagory == "mutation":
                    name = mutation
 
                if mutation not in mutations:
                    mutations[mutation] = Entry(splitline[4], splitline[5], splitline[6], name, splitline[7])
                    if catagory == "mutation": 
                        mutations[mutation].diseases=set([])
                    mutations[mutation].weight = 1
                else:
                    mutations[mutation].weight += 1
                if catagory == "mutation":
                    mutations[mutation].diseases.add(disease)

        gr = genomicrange()
        for entry in mutations.values():
            if catagory == "mutation":
                entry.diseases = "; ".join(sorted(entry.diseases))
            gr.construct(entry)
        gr.sort()
        return gr


    @classmethod
    def load_vcf(genomicrange, path, range_of_interest=None): #range_of_interest = genomic range covering variants to be loaded
        gr = genomicrange()
        if range_of_interest is not None:
            range_of_interest = range_of_interest.merged
        with file(path, "rU") as f:
            for line in f:
                if not line.startswith("#"):
                    splitline = line.split("\t")
                    chrom, untrimmedpos, snpid, untrimmedref, alts, qual, filters, junk = splitline[0:8]
                    if len(splitline) == 10:
                        headings, values = splitline[8:10]
                    assert len(splitline) == 8 or len(splitline) == 10
                    
                    if not chrom.startswith("chr"):
                        chrom = "chr{0}".format(chrom)
                    if chrom == "chr23":
                        chrom = "chrX"
                    elif chrom == "chr24":
                        chrom = "chrY"
                    elif chrom == "chr25":
                        chrom = "chrM"

                    quality = float(qual) if (qual!=".") else "."
                    passfilter = "." if filters == "." else all([code in ("PASS", "LowVariantFreq") for code in filters.split(";")])
#                    if filters != ".":
#                        passfilter = True
#                        for code in filters.split(";"):
#                            if code in ("LowGQX", "LowGQ"):
#                                passfilter = False
#                            elif code not in ("PASS", "LowVariantFreq", "R8"):
#                                passfilter = False
#                                print code
#                                #raise CoverMiException("Unknown filter code {0} in VCF {1}".format(code, path))
                    try:
                        alt_depths = [ float(depth) for depth in values.split(":")[headings.split(":").index("AD")].split(",") ]
                        total_depth = sum(alt_depths)
                    except (ValueError, UnboundLocalError):
                        pass

                    untrimmedpos = int(untrimmedpos)
                    alts = alts.split(",")
                    for alt_number in range(0, len(alts)):
                        alt = alts[alt_number]
                        ref = untrimmedref
                        start = untrimmedpos
                        while len(ref) > 1 and len(alt) > 1:
                            if ref[0:2] == alt[0:2]:
                                ref = ref[1:]
                                alt = alt[1:]
                                start += 1
                            elif ref[-1] == alt[-1]:
                                ref = ref[:-1]
                                alt = alt[:-1]
                            else:
                                break
                        refalt_length_difference = len(ref) - len(alt)
                        if refalt_length_difference == 0:#snp
                            stop = start
                        elif refalt_length_difference < 0:#alt longer than ref therefore insertion
                            stop = start + 1
                        else:#deletion 
                            stop = start + 1 + refalt_length_difference

                        if range_of_interest is not None:
                            if chrom not in range_of_interest:
                                continue
                            covered = False
                            for roi_entry in range_of_interest[chrom]:
                                if roi_entry.start <= start and roi_entry.stop >= stop:
                                    covered = True
                                    break
                                if roi_entry.start > stop:
                                    break
                            if not covered:
                                continue

                        entry = Entry(chrom, start, stop, "{0}:{1} {2}>{3}".format(chrom, start, ref, alt), "+")#?????strand
                        entry.indel = refalt_length_difference != 0
                        if quality != ".":
                            entry.quality = quality
                        if passfilter != ".":
                            entry.passfilter = passfilter
                        try:
                            if total_depth == 0:
                                continue
                            else:
                                entry.vaf = alt_depths[alt_number+1]/total_depth
                        except UnboundLocalError:
                            pass
                        gr.construct(entry)
        gr.sort()
        return gr


    def construct(self, entry):################################# ?add
        if entry.chrom not in self:
            self[entry.chrom] = []
        self[entry.chrom].append(entry)
        return self
            

    def sort(self):
        for chrom in self:
            self[chrom].sort()


    @property
    def all_entries(self):
        for chrom in sorted(self, key=type(self).chromosome_number):
            for entry in self[chrom]:
                yield entry

    @property
    def merged(self): # The name of a range will be the name of the first sub-range that was merged into it
        merged = type(self)()
        for chrom in self:
            merged[chrom] = [copy.copy(self[chrom][0])]
            for entry in self[chrom]:
                if entry.start-1 <= merged[chrom][-1].stop:
                    merged[chrom][-1].stop = max(entry.stop, merged[chrom][-1].stop)
                else:
                    merged.construct(copy.copy(entry))
        merged.sort()
        return merged


    def extended_to_include_touching_amplicons(self, gr):
         extended = type(self)()
         amplicons = type(self)()########################################## = ?Gr
         for chrom in self:
             for a in self[chrom]:
                 new_entry = copy.copy(a)
                 if chrom in gr:
                     for b in gr[chrom]:
                         if b.start > a.stop:
                             break
                         if (a.start <= b.start <= a.stop) or (a.start <= b.stop <= a.stop) or b.stop > a.stop:
                             new_entry.start = min(new_entry.start, b.start)
                             new_entry.stop = max(new_entry.stop, b.stop)
                             amplicons.construct(copy.copy(b))
                 extended.construct(new_entry)
         return (extended, amplicons)


    def overlapped_by(self, gr):# If range gr contains overlapping ranges then we may get multiple copies of the overlapping ranges
         trimmed = type(self)()
         for chrom in self:
            if chrom in gr:
                for a in self[chrom]:
                    for b in gr[chrom]:
                        if b.start > a.stop:
                            break
                        if (a.start <= b.start <= a.stop) or (a.start <= b.stop <= a.stop) or b.stop > a.stop:
                            new_entry = copy.copy(a)
                            new_entry.start = max(a.start, b.start)
                            new_entry.stop = min(a.stop, b.stop)
                            trimmed.construct(new_entry)
         trimmed.sort() # Only needed if gr contains overlapping ranges which should not happen with sane use
         return trimmed


    def not_touched_by(self, gr):
        untouched = type(self)() 
        for chrom in self:
           for a in self[chrom]:
               touching = False
               if chrom in gr:
                   for b in gr[chrom]:
                       if b.start > a.stop:
                           break
                       if (a.start <= b.start <= a.stop) or (a.start <= b.stop <= a.stop) or b.stop > a.stop:
                           touching = True
                           break
               if not touching:
                   untouched.construct(copy.copy(a))
        untouched.sort()
        return untouched


    def touched_by(self, gr):
         touched = type(self)() 
         for chrom in self:
            for a in self[chrom]:
                if chrom in gr:
                    for b in gr[chrom]:
                        if b.start > a.stop:
                            break
                        if (a.start <= b.start <= a.stop) or (a.start <= b.stop <= a.stop) or b.stop > a.stop:
                            touched.construct(copy.copy(a))
                            break
         touched.sort()
         return touched

    
    def subranges_covered_by(self, gr):
         covered = type(self)()
         for chrom in self:
            if chrom in gr:
                for a in self[chrom]:
                    for b in gr[chrom]:
                        if b.start > a.start:
                            break
                        if b.stop >= a.stop:
                            covered.construct(copy.copy(a))
                            break
         covered.sort()
         return covered 


    def combined_with(self, other):
        combined = type(self)()
        for gr in (self, other):
            for chrom in gr:
                for entry in gr[chrom]:
                    combined.construct(copy.copy(entry))
        combined.sort()
        return combined


    def subset(self, names, exclude=False, genenames=False):
        if not hasattr(names, "__iter__"):
            names = [names]
        if genenames:
            names = [name.split()[0] for name in names]
        gr = type(self)()
        for entry in self.all_entries:
            if ((entry.name.split()[0] if genenames else entry.name) in names) ^ exclude:
                gr.construct(copy.copy(entry))
        return gr


    @property
    def names(self):
        names = set([])
        for entry in self.all_entries:
            names.add(entry.name)
        return sorted(names)


    @property
    def number_of_components(self):
        output = 0
        for chrom in self:
            output += len(self[chrom])
        return output


    @property
    def number_of_weighted_components(self):
        output = 0
        for entry in self.all_entries:
            output += entry.weight
        return output


    @property
    def base_count(self):
        output = 0
        for entry in self.all_entries:
            output += entry.stop - entry.start + 1
        return output


    @property
    def locations_as_string(self):
        output = []
        for entry in self.all_entries:
            output.append("{0}:{1}-{2}".format(entry.chrom, entry.start, entry.stop))
        return ", ".join(output)


    @property
    def names_as_string(self):   
        namedict = {}
        for entry in self.all_entries:
            if entry.name not in namedict:
                namedict[entry.name] = []
            try:
                namedict[entry.name].append(entry.exon)
            except AttributeError:
                pass
        namelist = []
        for name, numbers in namedict.items():
            numbers.sort()
            exons = []
            index = 0
            while index<=len(numbers)-1:
                start = numbers[index]
                stop = start
                for index2 in range(index+1, len(numbers)+1):
                    if index2 == len(numbers) or numbers[index2] != stop+1:
                        break
                    stop += 1
                exons.append("e{0}{1}".format(start, "" if (start==stop) else "-{0}".format(stop)))
                index = index2
            namelist.append("{0} {1}".format(name, ",".join(exons)))
            namelist.sort()
        return ", ".join(namelist).strip()####################?strip needed


    @property
    def is_empty(self):
        return len(self) == 0


    def save(self, f): #Save type(self)() object in bedfile format, START POSITION IS ZERO BASED
        try:
            for entry in self.all_entries:
                f.write("{0}\t{1}\t{2}\t{3}\t.\t{4}\n".format(entry.chrom, entry.start-1, entry.stop, entry.name, entry.strand))
        except AttributeError:
            with file(f, "wt") as realfile:
                self.save(realfile)
            




from pkg_resources import resource_string
import pdb
from gr import Gr
import subprocess, os
try:
    import pyper
except ImportError:
    pass

class _Encode(object):
    def __init__(self, name, strand, adj):
        self.name = name
        self.strand = strand
        self.adj = adj

    def encode(self, x, y, component):
        for beginning, end, fixed, scaling in self.adj:
            if x >= beginning and x <= end:
                break
        rel_pos = x - beginning
        adjustment = fixed + (rel_pos - (rel_pos/scaling))
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n".format(x, y, self.name, component, self.strand, adjustment)


def plot(coverage, panel, outputstem):
    if "Transcripts" not in panel:
        return

    output_file = outputstem+"_plot.pdf"
    plotname = panel["Filenames"]["Sample"] if ("Sample" in panel["Filenames"]) else panel["Filenames"]["Panel"]
    rdataframe = outputstem+"_Rdata.tsv"
    with file(rdataframe, "wt") as f:
        s = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\n"
        f.write(s.format("x", "y", "name", "component", "strand", "adjustment"))

        for entry in panel["Transcripts"].all_entries:
                chrom = entry.chrom
                name = entry.name
                strand = entry.strand
                gr1 = Gr(entry)

                if "Amplicons" in panel:
                    gr1, amplicons = gr1.extended_to_include_touching_amplicons(panel["Amplicons"])
                else:
                    amplicons = Gr()
                auc = 0
                for amplicon in amplicons.all_entries:
                    auc += amplicon.stop - amplicon.start + 1                
                INTRON_SIZE = auc / amplicons.number_of_components if (amplicons.number_of_components>0) else 200

                start = gr1[chrom][0].start
                stop = gr1[chrom][0].stop
                
                exons = panel["Exons"].touched_by(gr1).subset(name) if ("Exons" in panel) else Gr()

                blocks = amplicons.combined_with(exons).merged
                # adj = [start, stop, fixed subtraction, scaling factor]
                adj = []
                end_of_prev_block = start-2
                fixed_total = 0
                for bentry in blocks.all_entries:
                    bstart = bentry.start
                    bstop = bentry.stop
                    intron_size = bstart - end_of_prev_block - 1
                    adj.append((end_of_prev_block+1, bstart-1, fixed_total, max(float(intron_size)/INTRON_SIZE, 1)))
                    fixed_total += max(intron_size - INTRON_SIZE, 0)
                    adj.append((bstart, bstop, fixed_total, 1))
                    end_of_prev_block = bstop
                adj.append((end_of_prev_block+1, stop+1, fixed_total, max(float(stop - end_of_prev_block)/INTRON_SIZE, 1) if (not blocks.is_empty) else 1))

                line = _Encode(name, strand, adj)

                if "Amplicons" in panel:
                    for aentry in amplicons.all_entries:
                        f.write(line.encode(aentry.start, aentry.stop, "amplicon"))
                   
                f.write(line.encode(start-1, 0, "coverage"))
                for cstart, cstop, cdepth in coverage[chrom]:
                    if cstop >= start:
                        f.write(line.encode(max(start, cstart), cdepth, "coverage"))
                        if cstop >= stop:
                            f.write(line.encode(min(stop, cstop), cdepth, "coverage"))
                            break
                f.write(line.encode(stop+1, 0, "coverage"))

                if "Exons" in panel:
                    for eentry in exons.all_entries:
                        f.write(line.encode(eentry.start, eentry.stop, "exon"))
                        f.write(line.encode((eentry.start+eentry.stop)/2, str(eentry.exon), "exon_number"))
        
                if "Variants_Mutation" in panel:
                    variants = gr1.overlapped_by(panel["Variants_Mutation"])
                    vycoord = panel["Options"]["Depth"] if ("Depth" in panel["Options"]) else 0
                    for ventry in variants.all_entries:
                            f.write(line.encode((ventry.start+ventry.stop)/2, vycoord, "variants"))
        
                if "Depth" in panel["Options"]:
                    f.write(line.encode(start, panel["Options"]["Depth"], "minimum"))

                #if "Other" in panel:
                #    for oentry in gr1.overlapped_by(panel["Other"]).all_entries:
                #        f.write(line.encode(oentry.start, oentry.stop, "other"))

    genericrcode = resource_string(__name__, "covermiplot.R")
    rcode = genericrcode.replace("OUTPUT", os.path.abspath(output_file)).replace("INPUT", os.path.abspath(rdataframe)).replace("NAME", plotname).replace("\\", "/")
    r = pyper.R()
    r.run(rcode)    
    os.unlink(rdataframe)





import pkg_resources
try:
    from covermiglobals import *
    import GenomicRange 
except ImportError:
    from gr import Gr as GenomicRange
    VERSION = pkg_resources.require("CoverMi")[0].version

import time, os, pdb


def create(coverage, panel, output_path):

    try:
        output_file = os.path.join(output_path, panel["Filenames"]["Output"]+"_covermi_test_report.txt")
    except KeyError:
        output_file = output_path+"_covermi_test_report.txt"
    with file(output_file, "wt") as f:
        f.write("{0}{1}\n".format("CoverMi version:", VERSION))
        
        try:
            minimum_depth = panel["Depth"]
        except KeyError:
            minimum_depth = panel["Options"]["Depth"]

        targeted_exons = panel["Exons"] if ("Exons" in panel) else panel["Transcripts"]
        if "Amplicons" in panel:
            targeted_range = panel["Amplicons"].merged
            targeted_exons = targeted_exons.overlapped_by(targeted_range)
        else:
            targeted_range = panel["Transcripts"]
            targeted_exons = targeted_exons

        try:
            i = coverage.info(targeted_exons.renamed("all"), minimum_depth)[0]
        except AttributeError:
            i = coverage.calculate(targeted_exons, minimum_depth, total=True)
        f.write("\n{0} of {1} bases ({2:0.0f}%) covered at depth >= {3}  (mean depth {4})\n".format(
                i.bases_covered, i.bases, i.percent_covered, minimum_depth, i.depth_covered))

        f.write("\nCoverage by Gene (depth >= {0})\n".format(minimum_depth))
        try:
            info = coverage.info(targeted_exons.by_transcript, minimum_depth)
        except AttributeError:
            info = coverage.calculate(targeted_exons, minimum_depth)
        for i in info:
            f.write("{0:65}{1: >3.0f}% covered (mean depth {2})\n".format(i.name.split(" ")[0], i.percent_covered, i.depth_covered))

        f.write("\nExons with incomplete coverage at depth >= {0}\n".format(minimum_depth))
        try:
            info = coverage.info(targeted_exons, minimum_depth)
        except AttributeError:
            info = coverage.calculate(targeted_exons, minimum_depth, exons=True)
        for i in info:
            if i.incompletely_covered:
                mean_depth = "(mean depth {0})".format(i.depth_covered) if (i.bases_covered>0) else ""
                f.write("{0:65}{1: >3.0f}% covered {2}\n".format(i.name.split(" ")[0], i.percent_covered, mean_depth))
    
        f.write("\nRanges not covered at  depth >= {0}\n".format(minimum_depth))
        for i in info:
            if i.incompletely_covered:
                name = i.name.split(" ")
                name = name[0] + " " + name[-1]
                f.write("{0:65}{1} (mean depth {2})\n".format(name, i.range_uncovered.locations_as_string, i.depth_uncovered))

        if "Variants" in panel or "Variants_Gene" in panel:
            header = "\nKnown mutations covered at depth >= {0} by gene\n".format(minimum_depth)
            try:
                all_variants = panel["Variants"].by_gene
            except KeyError:
                all_variants = panel["Variants_Gene"]
            targeted_variants = all_variants.overlapped_by(targeted_range)
            try:
                info = coverage.info(all_variants, minimum_depth)
            except AttributeError:
                info = coverage.calculate(all_variants, minimum_depth)
            for i in info:
                mean_depth = "(mean depth {0})".format(i.depth_covered) if (i.components_covered>0) else ""
                detectable = targeted_variants.subset(i.name).number_of_components
                f.write("{0}{1:65}{2: >3} of {6: >3} ({5: >3.0f}%) detectable mutations ({4: >3.0f}% of {3: >3} known mutations) covered {7}\n".format(
                        header, i.name, i.components_covered, i.components, i.percent_components_covered, i.components_covered*100/max(detectable,1), detectable, mean_depth))
                header = ""


        if "Variants" in panel or "Variants_Disease" in panel:
            header = "\nKnown mutations covered at depth >= {0} by disease\n".format(minimum_depth)
            try:
                all_variants = panel["Variants"].by_disease
            except KeyError:
                all_variants = panel["Variants_Disease"]

            targeted_variants = all_variants.overlapped_by(targeted_range)
            try:
                info = coverage.info(all_variants, minimum_depth)
            except AttributeError:
                info = coverage.calculate(all_variants, minimum_depth)
            for i in info:

                mean_depth = "(mean depth {0})".format(i.depth_covered) if (i.components_covered>0) else ""
                try:
                    detectable = targeted_variants.subset(i.name, exactmatch=True).number_of_components
                except AttributeError:
                    detectable = targeted_variants.subset(i.name).number_of_components
                f.write("{0}{1:65}{2: >3} of {6: >3} ({5: >3.0f}%) detectable mutations ({4: >3.0f}% of {3: >3} known mutations) covered {7}\n".format(
                        header, i.name, i.components_covered, i.components, i.percent_components_covered, i.components_covered*100/max(detectable,1), detectable, mean_depth))
                header = ""

        if "Variants" in panel or "Variants_Mutation" in panel:
            header = "\nDetectable mutations not covered at depth >= {0}\n".format(minimum_depth)
            body = []
            try:
                info = coverage.info(panel["Variants"].overlapped_by(targeted_range), minimum_depth, combine=False)
            except AttributeError:
                info = coverage.calculate(panel["Variants_Mutation"].overlapped_by(targeted_range), minimum_depth)
            for i in info:
                if i.incompletely_covered:
                    try:
                        body += ["{0}{1:10}{2:30}{3:40}({4}) (mean depth {5})\n".format(
                            header, i.gene, i.range_uncovered.locations_as_string, i.mutation, i.disease, i.depth_uncovered))]
                    except AttributeError:
                        body += ["{0}{1:10}{2:30}{3:40}({4}) (mean depth {5})\n".format(
                            header, i.name.split()[0], i.range_uncovered.locations_as_string, i.name.split()[1], i.diseases, i.depth_uncovered))]
            
            if body:
                f.write(header)
                for line in sorted(body):
                    f.write(line)







